//
//  Excavator.swift
//  yumbo
//
//  Created by kunii on 2018/09/19.
//  Copyright © 2018年 Takahiro Kunii. All rights reserved.
//

import Foundation
import SceneKit

/// 油圧ショベルモデル
class Excavator {
    /// ノード識別用文字列
    private static let sholderID = "shoulder"  //  肩
    private static let boomID = "boom" //  ブーム
    private static let armID = "arm"   //  アーム
    private static let bucketID = "bucket" //  バケット
    
    /// 回転角の下限と上限(360度単位)
    private struct Limit {
        /// 下限
        let min:Float
        /// 上限
        let max:Float
    }

    /// 各ノードごとの回転角の下限と上限
    private static let limits = [
        sholderID:Limit(min:-180, max:180),
        boomID:Limit(min:0, max:90),
        armID:Limit(min:0, max:90),
        bucketID:Limit(min:-30, max:90)
    ]

    ///  回転速度は固定 （ラジアン/秒）
    private static let velocity = Float(10).radian

    /// 油圧ショベルのルートノード
    var root:SCNNode!
    
    ///  油圧ショベルのノード作成
    init() {
        //  土台
        let base = SCNBox()
        let baseNode = SCNNode(geometry: base)
        
        //  肩
        let shoulder = SCNSphere()
        let shoulderNode = SCNNode(geometry: shoulder)
        shoulderNode.name = Excavator.sholderID  //  検索できるように名前をつける
        baseNode.addChildNode(shoulderNode)
        
        guard let yumboURL = Bundle.main.url(forResource: "yumbo", withExtension: "dae") else {
            return
        }
        guard let yumboScene = try? SCNScene(url:yumboURL) else {
            return
        }
        
        //  ブーム
        guard let boomNode = yumboScene.rootNode.childNode(withName: Excavator.boomID, recursively: true) else {
            return
        }
        shoulderNode.addChildNode(boomNode)
        
        //  アーム
        guard let armNode = yumboScene.rootNode.childNode(withName: Excavator.armID, recursively: true) else {
            return
        }
        boomNode.addChildNode(armNode)
        
        //  バケット
        guard let bucketNode = yumboScene.rootNode.childNode(withName: Excavator.bucketID, recursively: true) else {
            return
        }
        armNode.addChildNode(bucketNode)
        
        //  土台はpositionの示す位置が底辺になるようにする
        baseNode.pivot = SCNMatrix4MakeTranslation(0, baseNode.geometry!.boundingBox.min.y, 0)
        
        //  肩である球体は土台の最上部が、球の中心位置になるように配置
        shoulderNode.position = SCNVector3(x:0, y:baseNode.geometry!.boundingBox.max.y, z:0)
        
        //  ブームはpositionの示す位置が底辺になるようにし、肩の最上部 - 30cm に配置（ちょっとめり込ます）
        boomNode.position = SCNVector3(x:0, y:shoulderNode.geometry!.boundingBox.max.y - 0.3, z:0)
        
        //  アームはpositionの示す位置が底辺になるようにし、ブームの最上部に配置
        armNode.position = SCNVector3(x:0, y:boomNode.geometry!.boundingBox.max.y, z:0)
        
        //  バケットはアームの最上部に配置
        bucketNode.position = SCNVector3(x:0, y:armNode.geometry!.boundingBox.max.y, z:0)
        
        base.firstMaterial?.diffuse.contents = UIColor.red
        shoulder.firstMaterial?.diffuse.contents = UIColor.yellow

        root = baseNode
    }
    
    /// 肩を回転させる
    ///
    /// - Parameter direction: -1：左回転,1：右回転
    func moveShoulder(direction:Float) {
        moveNode(name:Excavator.sholderID, direction:direction)
    }
    
    func moveBoom(direction:Float) {
        moveNode(name:Excavator.boomID, direction:direction)
    }
    
    func moveArm(direction:Float) {
        moveNode(name:Excavator.armID, direction:direction)
    }
    
    func moveBucket(direction:Float) {
        moveNode(name:Excavator.bucketID, direction:direction)
    }

    /// ノードの回転
    ///
    /// sholderIDノードのみy軸での回転となる
    ///
    /// - Parameters:
    ///   - name: ノードの指定
    ///   - direction: 回転方向
    private func moveNode(name:String, direction:Float) {
        //  指定ノードの回転角の上限下限を得る
        guard let limit = Excavator.limits[name] else {
            return
        }
        if let node = root.childNode(withName: name, recursively: true) {
            let animationName = name + " move"  //  アニメーション名決定
            
            //  まず停止：画面上の角度を設定することでアニメーション中なら強制停止になる
            if name == Excavator.sholderID {
                node.eulerAngles.y = node.presentation.eulerAngles.y
            } else {
                node.eulerAngles.z = node.presentation.eulerAngles.z
            }
            
            //  アニメーションを直接作成したので、強制停止するには削除が必要
            node.removeAction(forKey: animationName)
            if direction != 0 {
                //  limit.min 〜　limit.max の範囲を可動範囲とする
                //  現在の角度から、範囲内の残り可動角を算出
                var left = direction < 0 ? limit.min.radian : limit.max.radian
                left -= name == Excavator.sholderID ? node.eulerAngles.y : node.eulerAngles.z
                if abs(left) > 0.0 {
                    //  可動範囲内なので動かす
                    
                    //  角度に必要な時間を、動作速度から割り出す
                    let duration = abs(left) / Excavator.velocity
                    let action:SCNAction
                    if name == Excavator.sholderID {
                        action = SCNAction.rotateBy(x: 0, y: CGFloat(left), z: 0, duration: TimeInterval(duration))
                    } else {
                        action = SCNAction.rotateBy(x: 0, y: 0, z: CGFloat(left), duration: TimeInterval(duration))
                    }
                    //  アニメーション登録
                    node.runAction(action, forKey: animationName)
                }
            }
        }
    }
}

// MARK: - 360度値をラジアン値に変換するコンピューティドプロパティの定義
extension Float {
    /// 自身の値を360度値と解釈して、これをラジアン値に変換した値を戻す
    var radian : Float {
        return 2.0 * .pi * self / 360.0
    }
}
